--[[
--  TankMastery Addon for World of Warcraft
--
--  Basic set of tools that should help raider tanks improve their playstyle.
--]]

local version = "0.85";
local isDebug = false;

local ADDON_NAME = "TankMastery";
local addon = LibStub("AceAddon-3.0"):NewAddon(ADDON_NAME, "LibSink-2.0", "AceConsole-3.0")
local L	= LibStub("AceLocale-3.0"):GetLocale(ADDON_NAME)
local media = LibStub("LibSharedMedia-3.0") 
local LQT = LibStub:GetLibrary( "LibQTip-1.0" )
local SOUND = media.MediaType.SOUND
local DATABROKER = LibStub:GetLibrary("LibDataBroker-1.1"):NewDataObject(ADDON_NAME, {
		type = "data source",
		text = L["addon_name"],
    icon = [[Interface\ICONS\Ability_Physical_Taunt]],
})


local chanNames = {
  "Disabled",
	"< Empty >",
	"< Empty >",
	"< Empty >",
	"< Empty >",
	"< Empty >",
	"< Empty >",
	"< Empty >",
	"< Empty >",
	"< Empty >",
}

local lastNinja = nil;

local defaults = {
	profile = {
		sinkoptions = { sink20OutputSink = "ChatFrame"; },
		sound = nil,
--    raidicon = false,
  	outputChannel = 1,
		targetTankerIcon = {
--	    enabled = true,
	    locked = false,
    	size = 40,
    	min = 0.75,
    	loc = {
	      x = 0,
      	y = 0,
      	point = "CENTER",
      	relPoint = "CENTER",
    		},
	  },
	  tankerIcon = true,
  },
	realm = {
		tauntStats = { ["*"] = { ["*"] = 0 } }
	}
}
local options = {
    name = L["addon_name"],
    desc = L["addon_desc"],
    type = "group",
    get = "GetValue",
    set = "SetValue", 
    handler = addon,
    args = {
	    	output = addon:GetSinkAce3OptionsDataTable(),
	
		    sound = {
				    name = L["sound_name"],
				    desc = L["sound_desc"],
						type = "select",
				    order = 10,
						dialogControl = 'LSM30_Sound',
						values = AceGUIWidgetLSMlists.sound,
				},
--[[	Disabled for now...	
				raidicon = {
				    name = L["raidicon_name"],
				    desc = L["raidicon_desc"],
				    type = "toggle",
				    order = 20,
				},				
--]]
				outputChannel = {
					type = "select",
					name = "Secondary Output Channel",
					width = "150",
					order = 40,
					values = chanNames,
				},				
				tankerIcon = {
				    name = L["tankericon_name"],
				    desc = L["tankericon_desc"],
				    type = "toggle",
				    width = "300",
				    order = 50,
				},								
				unlocktankerIcon = {
				    name = L["tankericonlock_name"],
				    desc = L["tankericonlock_desc"],
				    type = "execute",
				    width = "100",
				    order = 60,
				    func = function() 
								addon:LockTankerIcon(false);
				    		end,
				},						
    },
}
options.args.output.inline = true;
options.args.output.order = 30;

local taunts = {
	355,   -- Warrior: Taunt
	1161,  -- Warrior: Challenging Shout
	5209,  -- Druid: Challanging Roar
	6795,  -- Druid: Growl
	27270, -- VoidWalker: Torment Rank 7 
	31789, -- Paladin: Righteous Defense
	33700, -- Felguard: Anguish Rank 3
	33701, -- VoidWalker: Suffering Rank 6
	47984, -- VoidWalker: Torment Rank 8
	47989, -- VoidWalker: Suffering Rank 7
	47990, -- VoidWalker: Suffering Rank 8
	47993, -- Felguard: Anguish Rank 4
	56222, -- DK: Dark Command
	59671, -- Warlock: Challanging Howl
	62124, -- Paladin: Hand of Reckoning
}
local mocks = {	
	694,   -- Warrior: Mocking Blow
	49576, -- DK: Death Grip
	20736, -- Hunter: Distracting Shot
}
local knockbacks = {
	53227, -- Druid: Typhoon
	59159, -- Shaman: Thunderstorm Rank 4
}
local tankAuras = {
	71,    -- Warrior: Defensive Stance
	9634,  -- Druid: Dire Bear Form
	25780, -- Paladin: Righteous Fury
	48263, -- Death Knight: Frost Presence
}
local COMBATLOG_OBJECT_TYPE_MAINTANK = 0x00040000;
local COMBATLOG_OBJECT_TYPE_PLAYER   = 0x00000400;

local backlogUnits = {
}
local buffRighteousFury = GetSpellInfo(25780)
local buffFrostPresence = GetSpellInfo(48263)
local blizzTanks = {
}
local watchedTanks = {
}
local ICON_LIST = {  
	"{star}",  
	"{circle}",  
	"{diamond}",  
	"{triangle}",  
	"{moon}",  
	"{square}",  
	"{cross}",  
	"{skull}",
}
-- Tooltip templates
local tooltip
local function Color(c, t) return "|cff"..c..t.."|r" end
local function HGreen(t) return Color("80ff80", t) end
local function HYellow(t) return Color("ffffa0", t) end
local function Green(t) return Color("80ff80", t) end
local function Blue(t) return Color("8080ff", t) end
local function Red(t) return Color("ff8080", t) end
local function WhiteParens(t) return "|cffffffff(|r"..t.."|cffffffff)|r" end

-- Addon MasterSwitch!
local isTankMasteryEnabled = false;
local isMonitoringTaunts = false;
local isInCombat = false;

function addon:OnInitialize()
	self.db = LibStub("AceDB-3.0"):New("TankMasteryDB", defaults, "Default")
	self:SetSinkStorage(self.db.profile.sinkoptions)
	local acreg = LibStub("AceConfigRegistry-3.0")
	acreg:RegisterOptionsTable(ADDON_NAME, options)
	local acdia = LibStub("AceConfigDialog-3.0")	
	acdia:AddToBlizOptions(ADDON_NAME, ADDON_NAME)
	addon:RegisterChatCommand(ADDON_NAME:lower(), function() acdia:Open(ADDON_NAME) end)
	
--	self.db:RegisterDefaults(
--	{
--	})
end

function addon:OnEnable()	
		-- only register events if we can tank... 			   
  local isTank = self:isPossibleTank("player");
  if (isTank > 0) then
	  		isTankMasteryEnabled = true
				TankMasteryFrame:UnregisterAllEvents();
	  		media.UnregisterAllCallbacks(self);
				
				TankMasteryFrame:RegisterEvent("COMBAT_LOG_EVENT_UNFILTERED");
				
				-- selective registering (optimization)
				if (isTank == 1) then
					TankMasteryFrame:RegisterEvent("UPDATE_SHAPESHIFT_FORM");
				elseif (isTank == 2) then
					TankMasteryFrame:RegisterEvent("UNIT_AURA");
				end			

				-- Leaving combat (for clearing up backlogs etc)
				TankMasteryFrame:RegisterEvent("PLAYER_REGEN_ENABLED"); 
				TankMasteryFrame:RegisterEvent("PLAYER_REGEN_DISABLED"); 
				TankMasteryFrame:RegisterEvent("PARTY_MEMBERS_CHANGED");
				TankMasteryFrame:RegisterEvent("RAID_ROSTER_UPDATE");				

				-- channel listing...
				TankMasteryFrame:RegisterEvent("CHAT_MSG_CHANNEL_NOTICE");
				addon:populateChanNames(); -- once to support reloadui as well...
									
				-- TargetTankingWatch Icon
				TankMasteryFrame:RegisterEvent("UNIT_TARGET");	
					
				ChatFrame1:AddMessage("|cffffff00TankMastery Loaded (v".. version ..")...");
												
				-- The Big Watch butotn on the screen...
				self:initWatchIcon();
				
				-- startup init
				self:updateSelfTankingMode();
	else
		isTankMasteryEnabled = false
		ChatFrame1:AddMessage("|cffffff00TankMastery NOT Loaded. We are not a tanking class.");
	end	
	
	TankMasteryFrame:Hide();		
end

-- #INITIALIZE
function addon:initWatchIcon()
  self.watchIcon = CreateFrame("Button","TMWatchImage",UIParent);	
	self.watchIcon.anchor = CreateFrame("Frame","TMWatchAnchor",UIParent);
  self.watchIcon.anchor:ClearAllPoints();
  self.watchIcon.anchor:SetWidth(30);
  self.watchIcon.anchor:SetHeight(30);
  self.watchIcon.anchor:SetBackdrop({
		bgFile = "Interface\\Tooltips\\UI-Tooltip-Background",
		edgeFile="Interface\\Tooltips\\UI-Tooltip-Border",
		tile = true,
		tileSize = 16,
		edgeSize = 16,
		insets = {
			left = 5,
			right = 5,
			top = 5,
			bottom = 5
		},
	});
	self.watchIcon.anchor:SetBackdropColor(0.1,0.1,0.3)
  local ploc = self.db.profile.targetTankerIcon.loc;
  self.watchIcon.anchor:EnableMouse(true);
  self.watchIcon.anchor:SetMovable(true);
  self.watchIcon.anchor:SetPoint( ploc.point, UIParent, ploc.relPoint, ploc.x, ploc.y );
  self.watchIcon.anchor:SetFrameLevel(5);
  self.watchIcon.anchor:RegisterForDrag("LeftButton");
  self.watchIcon.anchor:SetScript("OnDragStart", function() self.watchIcon.anchor:StartMoving(); end );
  self.watchIcon.anchor:SetScript("OnDragStop", function() self.watchIcon.anchor:StopMovingOrSizing(); local _;
    ploc.point, _, ploc.relPoint, ploc.x, ploc.y = self.watchIcon.anchor:GetPoint(); end);
  self.watchIcon:ClearAllPoints();
  self.watchIcon:SetPoint("CENTER", self.watchIcon.anchor, "CENTER", 0, 0);
  self.watchIcon:SetHeight(self.db.profile.targetTankerIcon.size)
  self.watchIcon:SetWidth(self.db.profile.targetTankerIcon.size)
  self.watchIcon:EnableMouse(false)
  self.watchIcon.animating = false;
  self.watchIcon:SetResizable(true);
  self.watchIcon:SetMaxResize(300,300);
  self.watchIcon:SetMinResize(50,50);  
  if self.db.profile.tankerIcon then
	  if self.db.profile.targetTankerIcon.locked then
	    self.watchIcon:Hide();
	  else
	    self.watchIcon:Show();
	  end  
	else
		 self.watchIcon:Hide();
	end
	  
	self:setTankerIcon(1);
	  
	self.watchIcon.anchor:SetScript("OnEnter",
		function()
			GameTooltip:SetOwner(this, "ANCHOR_BOTTOMRIGHT")
			GameTooltip:AddLine(L["tt_tankericon1L"])
			GameTooltip:AddLine(L["tt_tankericon2L"])
			GameTooltip:AddLine(L["tt_tankericon3L"])
			GameTooltip:Show()
		end
	);
  
  self.watchIcon.anchor:SetScript("OnLeave", function() GameTooltip:Hide(); end);  
	self.watchIcon.anchor:SetScript("OnMouseUp",
		function()
			if IsShiftKeyDown() then
				self:LockTankerIcon(true);
			end
		end
	);
  
  self:LockTankerIcon(self.db.profile.targetTankerIcon.locked);
end
function addon:setTankerIcon(mode)
	if (mode == 0) then
		self.watchIcon:Hide();
		return;
	end 
	
	self.watchIcon:Show();
	if (mode == 1) then -- self
		self.watchIcon:SetNormalTexture("Interface\\Icons\\INV_Shield_32");
	elseif (mode == 2) then -- another tank, and tank NOT targeting him
		self.watchIcon:SetNormalTexture("Interface\\Icons\\INV_Shield_09");
	elseif (mode == 3) then -- another tank, and tank targeting him
		self.watchIcon:SetNormalTexture("Interface\\Icons\\INV_Shield_06");
	elseif (mode == 4) then -- a player
		self.watchIcon:SetNormalTexture("Interface\\Icons\\Spell_Shadow_MindSteal");
	elseif (mode == 5) then -- another mob
		self.watchIcon:SetNormalTexture("Interface\\Icons\\Spell_Shadow_NetherCloak");
	else
		self.watchIcon:SetNormalTexture("Interface\\Icons\\Spell_Holy_Power");
	end
end
function addon:LockTankerIcon(locked)
  if (locked) then
    self.watchIcon.anchor:Hide();
    self.watchIcon:Hide();
    self.db.profile.targetTankerIcon.locked = true;
  else
    self.watchIcon.anchor:Show();
    self.watchIcon:Show();
    self.db.profile.targetTankerIcon.locked = false;
  end    
end

function addon:populateChanNames()
    for i=1,9 do
        channelNum, channelName = GetChannelName(i);
        if (channelNum > 0 and channelName ~= nil) then
        	if (string.sub(chanNames[i + 1], 1, 1) == "<") then
            chanNames[i + 1] =  i..". ".. channelName;
          end
        end
    end
end
function addon:getRaidIndex(name) 
	for i = 1, GetNumRaidMembers() do 
		if UnitName("raid"..i) == name then 
			return i 
		end 
	end 
end
function addon:isRaidMT(name)
	name, rank, subgroup, level, class, fileName, zone, online, isDead, role, isML = GetRaidRosterInfo(self:getRaidIndex(name));
	
	if (role == "MAINTANK") then
		return 1;
	else
		return 0;
	end
end
function addon:initBlizzTanks()
	if UnitInRaid("player") then
		self:debugPrint("Initializing Blizzard MTs...");
		wipe(blizzTanks);
		for i = 1,GetNumRaidMembers() do
		   local name, rank, subgroup, level, class, fileName, zone, online, isDead, role, isML = GetRaidRosterInfo(i);
		 
		   if (role) then
			   if (role == "MAINTANK") then
			   	self:debugPrint("* MT: " .. UnitName("raid" .. i));
		   		self:setPlayerIsTanking(blizzTanks, UnitGUID("raid" .. i), 0);
		   	end
		   end
		end
	end
end

function addon:preCombat()
	self:debugPrint("Pre-combat...");
end
function addon:afterCombat()
	self:debugPrint("After-combat...");
	wipe(backlogUnits);
end
function addon:cleanWatchedTanks()
	self:debugPrint("Cleaning watched-tank list...");
	wipe(watchedTanks);
	wipe(blizzTanks);
end
function addon:trimWatchedTanks()
		if (addon:inGroup("player") == false) then
			addon:cleanWatchedTanks();
		else
			if UnitInRaid("player") then
--				for i = 1,GetNumRaidMembers() do
--				   UnitName("raid"..i)
--				end
			elseif UnitInParty("player") then
--				for i = 1,GetNumPartyMembers() do
--				   UnitName("raid"..i)
--				end				
			end
		end
end

function TankMasteryFrame_OnEvent(event, ...)
	-- deduce event type	
	if (event == "UPDATE_SHAPESHIFT_FORM") then
		addon:updateSelfTankingMode();
	elseif (event == "UNIT_AURA" and arg1 == "player") then
		addon:updateSelfTankingMode();
	elseif (event == "UNIT_TARGET" and addon.db.profile.tankerIcon and (arg1 == "player" or arg1 == "target")) then
		if UnitGUID("target") then	    
			  -- targetfriendly check as well... we want to work on only unfriendly targets...
			  if ((not (UnitIsFriend("player","target"))) and UnitGUID("targettarget") and UnitGUID("targettargettarget") and UnitGUID("player") ~= UnitGUID("target")) then
					if (UnitGUID("player") == UnitGUID("targettarget")) then -- self
						addon:setTankerIcon(1);
					elseif ((addon:TableContainsValue(blizzTanks, UnitGUID("targettarget"))) or (addon:TableContainsValue(watchedTanks, UnitGUID("targettarget")))) then -- TargetTarget is Raid MT or another deduced tank
						if (UnitGUID("targettargettarget") == UnitGUID("target")) then -- is mob's target targeting him back?
							addon:setTankerIcon(3);
						else
							addon:setTankerIcon(2);
						end
					elseif (addon:inGroup(UnitName("targettarget"))) then -- TargetTarget is in our Raid/Party
						addon:setTankerIcon(4);
					else -- everything else...
						addon:setTankerIcon(999);
					end;
			  else
					addon:setTankerIcon(0);
			  end
		else
			addon:setTankerIcon(0);
		end
	elseif (event == "PLAYER_REGEN_ENABLED") then
		isInCombat = false;
		addon:afterCombat();
	elseif (event == "PLAYER_REGEN_DISABLED") then
		isInCombat = true;
		addon:preCombat();
	elseif (event == "PARTY_MEMBERS_CHANGED") then
		addon:trimWatchedTanks();
	elseif (event == "RAID_ROSTER_UPDATE") then
		addon:trimWatchedTanks();
		addon:initBlizzTanks();
	elseif (event == "CHAT_MSG_CHANNEL_NOTICE") then
		if (arg7 == 0) then
			chanNames[tonumber(arg8) + 1] = arg4;
		else
			chanNames[tonumber(arg8) + 1] = arg4 .. " (!!!)";
		end
  elseif (event == "COMBAT_LOG_EVENT_UNFILTERED") then
 		if (arg4 ~= "" and addon:inGroup(arg4)) then
			addon:TankMasteryFrame_CombatLogEvent(event, ...);
		end
  end
end

function addon:TankMasteryFrame_CombatLogEvent(event, ...)
	local timestamp, ttype, srcGUID, srcName, srcFlags, dstGUID, dstName, dstFlags, spellid, spellname, spellschool, pwrAmount, pwrType = select(1, ...);
	local tauntmsg = "";
	local isAOE = false;
	local targetName = "";	

	if (ttype == "SPELL_CAST_SUCCESS") then			
		-- Taunt/Mock/Knockback checks...		
		if (UnitGUID("targettarget")) and (UnitGUID("target")) and (UnitGUID("targettarget") == UnitGUID("player")) 
			and (UnitGUID("player") ~= srcGUID or isDebug)
			and (not CombatLog_Object_IsA(dstFlags, COMBATLOG_OBJECT_TYPE_PLAYER)) then			
				if ((dstName == nil or dstName == "") and UnitName("target")) then			
					--self:debugPrint("Non-targeted spell: ".. spellname);
					dstName = UnitName("target");
					isAOE = true;
				end
				targetName = dstName;
				
				if (addon:inGroup("player")) then
			  	  if GetRaidTargetIndex("target") then
				  		targetName = ICON_LIST[GetRaidTargetIndex("target")].. dstName;
				  	end
				end
				
				-- taunts etc
				if (self:TableContainsValue(taunts, spellid)) then
					if (not isAOE) then -- only my target
						if (UnitGUID("target") == dstGUID) then						
							tauntmsg = srcName.." taunted "..targetName.." ("..spellname..").";
						end
					else
						tauntmsg = srcName.." aoe-taunted "..targetName.." ("..spellname..").";
					end
				end
	
				-- mocks etc
				if (self:TableContainsValue(mocks, spellid)) then
					tauntmsg = srcName.." mocked "..targetName.." ("..spellname..").";
				end	
	
				-- knockback etc
				if (self:TableContainsValue(knockbacks, spellid)) then
					-- shammy Glyph check...
					if (spellid == 59159 and (self:TableContainsValue(backlogUnits, srcGUID)) == false) then
						self:debugPrint(srcName.." placed on backlog for knockback.");
						table.insert(backlogUnits, srcGUID);
					else
						tauntmsg = srcName.." knocked-away "..targetName.." ("..spellname..")";
					end				
				end				
		end
	elseif (ttype == "SPELL_MISSED" and srcGUID == UnitGUID("player")) then	      
	      if (self:TableContainsValue(taunts, spellid)) then
	      	local missType = arg12;
	      	tauntmsg = "! Failed to "..spellname.." mob "..dstName.." ("..missType..").";
	      end
	elseif (ttype == "SPELL_AURA_APPLIED" and srcGUID ~= UnitGUID("player")) then	
		if (self:TableContainsValue(tankAuras, spellid)) then
			self:debugPrint(srcName.." is also tanking now.");
			self:setPlayerIsTanking(watchedTanks, srcGUID, 0);
		end
	elseif (ttype == "SPELL_AURA_REMOVED" and srcGUID ~= UnitGUID("player")) then
		if (self:TableContainsValue(tankAuras, spellid)) then
			self:debugPrint(srcName.." is NOT tanking now.");
			self:setPlayerIsTanking(watchedTanks, srcGUID, 1);
		end		
	elseif (ttype == "SPELL_ENERGIZE" and (spellid == 59159)) then -- Backlog for Knockback... Checkign mana returned
		self:debugPrint(srcName.." is Energized for knockback.");
		if (self:TableContainsValue(backlogUnits, srcGUID)) then
			local power = UnitPower(srcName, 0);
			self:debugPrint(srcName.." mana returned = " .. pwrAmount);
			self:debugPrint(srcName.." total mana = " .. power);
			if (power > 0) then
				-- check if its 8% or 10% mana returned back
				if (power * 0.09 > pwrAmount) then
					tauntmsg = srcName.." knocked-away "..targetName.." using "..spellname..".";
				end				
			end
			self:tremovebyval(backlogUnits, srcGUID);
		end
	end	
		
	if (tauntmsg ~= "" and isMonitoringTaunts == true) then
--				  tauntStats = self.db.realm.tauntStats
	
--			    if self.db.profile.raidicon then 
--			        if UnitInRaid("player") and (IsRaidLeader() or IsRaidOfficer()) then
--			        	if not GetRaidTargetIndex(name) then
--			        		SetRaidTarget(name, ICON_TAG_LIST[ICON_TAG_RAID_TARGET_MOON1:lower()])
--			        		self.lastplayer = name
--			        	end
--			        end
--			    end				
	  addon:Pour(tauntmsg);
	  
	  outChan = self.db.profile.outputChannel;
	  if (outChan > 1 and chanNames[outChan] ~= "" and (string.sub(chanNames[outChan], 1, 1) ~= "<")) then
	  	SendChatMessage(tauntmsg, "CHANNEL", nil, chanNames[outChan]);
	  end
	  
  	if addon.db.profile.sound then
			PlaySoundFile(media:Fetch(SOUND, addon.db.profile.sound))
		end
	end
end

local function toggleOptionsWindow()
  local acd = LibStub("AceConfigDialog-3.0")
  if acd.OpenFrames[ADDON_NAME] then
      acd:Close(ADDON_NAME)
  else
      acd:Open(ADDON_NAME)
  end
end
function DATABROKER.OnClick(frame, button)
	if button == "RightButton" then
        toggleOptionsWindow()
	end
end
function DATABROKER.OnEnter(self)
  tooltip = LQT:Acquire(ADDON_NAME, 2, "LEFT", "RIGHT")
  tooltip:Clear()
  if (isTankMasteryEnabled == true) then
  	tooltip:AddHeader(HGreen(ADDON_NAME).." "..Blue("v"..version));
  else
  	tooltip:AddHeader(Red(ADDON_NAME));
  end
	tooltip:AddLine(L["tt_clickhint3L"].." "..L["tt_clickhint3R"])
	tooltip:EnableMouse()
	tooltip:SmartAnchorTo( self )
  tooltip:Show()
end
function DATABROKER.OnLeave(self)
    tooltip = LQT:Release(tooltip)
end

-- #FUNCTIONS
function addon:ClearNinjaIcon()
--[[
    if (addon.lastNinja) then 
        SetRaidTarget(self.lastNinja, 0)
        addon.lastNinja = nil
    end
--]]  
end

function addon:debugPrint(msg)
	if (isDebug == true) then
		ChatFrame1:AddMessage("DEBUG : |cFF00FF00" .. msg);
	end	
end

function addon:TableContainsValue(table, value)
      if (table and value) then
              for _, v in ipairs(table) do
                      if (v == value) then
                              return true;
                      end
              end
              return false;
      end
end

-- #TANKINGSTATUS
function addon:updateSelfTankingMode()
		if (self:isCurrentlyTanking("player") == true) then 
			self:debugPrint("Tanking...");
			isMonitoringTaunts = true;			
		else
			self:debugPrint("NOT Tanking...");
			isMonitoringTaunts = false;
			self:setTankerIcon(0);
	  end
end
function addon:isPossibleTank(sTarget)
	local localizedClass, PlayerClass = UnitClass(sTarget);
	if PlayerClass == "DRUID" then 
		return 1;
	end
	if PlayerClass == "WARRIOR" then 
		return 1;
	end
	if PlayerClass == "PALADIN" then 
		return 2;
	end
	if PlayerClass == "DEATHKNIGHT" then 
		return 1;
	end 
	return 0;
end
function addon:isCurrentlyTanking(sTarget)
	local localizedClass, PlayerClass = UnitClass(sTarget);
	-- Bear form
	if PlayerClass == "DRUID" and GetShapeshiftForm(true) == 1 then 
		return true;
	end
	-- Defensive stance
	if PlayerClass == "WARRIOR" and GetShapeshiftForm(true) == 2 then 
		return true;
	end
	-- Righteous Fury 
	if PlayerClass == "PALADIN" and UnitAura(sTarget, buffRighteousFury) then 
		return true;
	end
	-- Frost Presence  UnitBuff("player", buffFrostPresence) 
	if PlayerClass == "DEATHKNIGHT" and GetShapeshiftForm(false) == 2 then 
		return true;
	end

	return false;
end
function addon:inGroup(sTarget)
	if (UnitInParty(sTarget) or UnitInRaid(sTarget)) then
		return true;
	else
		return false;
	end
end
 -- tremovebyval: remove a table row given its value
function addon:tremovebyval(tab, val)
   for k,v in pairs(tab) do
     if(v==val) then
       table.remove(tab, k);
       return true;
     end
   end
   return false;
end
function addon:setPlayerIsTanking(whichList, sTarget, stat)
	if (stat == 0) then
		-- TankList Management.. ADDing
		if (self:TableContainsValue(whichList, sTarget) == false) then
			tinsert(whichList, sTarget);
		end
	else
		-- TankList Management.. REMOVEing
		if (self:TableContainsValue(whichList, sTarget) == true) then
			self:tremovebyval(whichList, sTarget);
		end
	end
end

-- #OPTIONS
function addon:SetValue(info, b)
	self.db.profile[info[#info]] = b		
end
function addon:GetValue(info)
	return self.db.profile[info[#info]]
end